library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity estec_lab_top is
  generic (
    dbcntmax:           integer range 10 to 1000000 := 25000;   -- 500 us @ 50 MHz
    cppb:               integer range 50 to 200000  := 433      -- 115.2 kbps @ 50 MHz
  );
  
  port (
    clk:               in  std_logic;
    rst_n:             in  std_logic;
    rx:                in  std_logic;
    tx:                out std_logic;
    transmitter_ready: out std_logic;
    parity_error_flag: out std_logic;
    btn_n:             in  std_logic;
    switches:          in  std_logic_vector (7 downto 0);
    ssd_digit_0:       out std_logic_vector (7 downto 0);
    ssd_digit_1:       out std_logic_vector (7 downto 0)
  );
end entity estec_lab_top;

architecture rtl of estec_lab_top is

  signal rst:                  std_logic;
  signal rst_d:                std_logic;
  signal rst_sync:             std_logic;
  signal btn:                  std_logic;
  signal btn_debounced:        std_logic;
  signal btn_debounced_d:      std_logic;
  signal btn_debounced_rising: std_logic;
  signal new_frame_received:   std_logic;
  signal frame_out:            std_logic_vector (7 downto 0);
  signal parity_error:         std_logic;
  signal ssd_digit_0_int:      std_logic_vector (7 downto 0);
  signal ssd_digit_1_int:      std_logic_vector (7 downto 0);

begin

  rst <= not rst_n;

  L_RESET_SYNC: process ( clk, rst )
  begin
    if rst = '1' then
      rst_d <= '1';
      rst_sync <= '1';
    elsif rising_edge(clk) then
      rst_d <= '0';
      rst_sync <= rst_d;
    end if;
  end process;
  
  ----------------------------------------
  
  btn <= not btn_n;
  
  L_DEBOUNCER:  entity work.debouncer(rtl)
            generic map (
              counter_max => dbcntmax
            )
            
            port map (
              clk        => clk,
              rst        => rst_sync,
              signal_in  => btn,
              signal_out => btn_debounced
            );
            
  process ( clk, rst_sync )
  begin
    if rst_sync = '1' then
      btn_debounced_d <= '0';
    elsif rising_edge(clk) then
      btn_debounced_d <= btn_debounced;
    end if;
  end process;
  btn_debounced_rising <= btn_debounced and not btn_debounced_d;
  
  ----------------------------------------
  
  L_UART_TRANSMITTER: entity work.uart_transmitter(rtl)
              generic map (
                clk_pulses_per_baud => cppb,
                data_bits           => 8,
                parity              => 2  -- odd
              )
              
              port map (
                clk         => clk,
                rst         => rst_sync,
                send_frame  => btn_debounced_rising,
                ready       => transmitter_ready,
                data_2_send => switches,
                tx          => tx
              );
  
  ----------------------------------------
  
  L_UART_RECEIVER:  entity work.uart_receiver(rtl)
              generic map (
                clk_pulses_per_baud => cppb,
                data_bits           => 8,
                parity              => 2  -- odd
              )
              
              port map (
                clk                => clk,
                rst                => rst_sync,
                rx                 => rx,
                new_frame_received => new_frame_received,
                frame_out          => frame_out,
                parity_error       => parity_error
              );
  
  ----------------------------------------
  
  L_PARITY_ERROR_FLAG: process ( clk, rst_sync )
  begin
    if rst_sync = '1' then
      parity_error_flag <= '0';
    elsif rising_edge(clk) then
      if new_frame_received = '1' then
        parity_error_flag <= parity_error;
      end if;
    end if;
  end process;
  
  ----------------------------------------
  
  L_SSD_ENCODER_DIGIT_0:  entity work.seven_segment_encoder(rtl)
                port map (
                  data_in  => frame_out(3 downto 0),
                  dp_off   => '1',
                  data_out => ssd_digit_0_int
                );
                
  L_SSD_ENCODER_DIGIT_1:  entity work.seven_segment_encoder(rtl)
                port map (
                  data_in  => frame_out(7 downto 4),
                  dp_off   => '1',
                  data_out => ssd_digit_1_int
                );
                
  L_SSD_OUT_FILTER: process ( clk, rst_sync )
  begin
    if rst_sync = '1' then
      ssd_digit_0 <= (others => '1');
      ssd_digit_1 <= (others => '1');
    elsif rising_edge(clk) then
      if new_frame_received = '1' then
        if parity_error = '0' then
          if unsigned(frame_out(7 downto 5)) > unsigned(frame_out(4 downto 0)) then
            ssd_digit_0 <= ssd_digit_0_int;
            ssd_digit_1 <= ssd_digit_1_int;
          end if;
        else
          ssd_digit_0 <= X"FF";
          ssd_digit_1 <= X"FF";
        end if;
      end if;
    end if;
  end process;
  
  ----------------------------------------

end architecture rtl;