library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.math_real.all;
----------------------------------------------------------------------------------------------------
package verification_package is

  procedure gen_uart_frame (
    data_2_send:            in  std_logic_vector;
    baud_delay:             in  time;
    parity:                 in  integer range 0 to 2;   -- 0: no parity; 1: odd, 2: even
    inject_parity_error:    in  boolean;
    baud_delay_uncertainty: in  time;
    signal tx:              out std_logic
  );

  procedure rec_uart_frame(
    baud_delay:             in  time;
    parity:                 in  integer range 0 to 2;   -- 0: no parity; 1: odd, 2: even
    timeout:                in  time;
    signal rx:              in  std_logic;
    signal parity_error:    out boolean;
    signal data_received:   out std_logic_vector
  );

  procedure bouncing_button(
    period:                 in  time;
    number:                 in  positive;
    signal button:          out std_logic
  );

  function seven_segment_decoder(
    data: std_logic_vector(7 downto 0)) 
  return std_logic_vector;
  
end package verification_package;
----------------------------------------------------------------------------------------------------
package body verification_package is
  
  procedure gen_uart_frame (
    data_2_send:            in  std_logic_vector;
    baud_delay:             in  time;
    parity:                 in  integer range 0 to 2;   -- 0: no parity; 1: odd, 2: even
    inject_parity_error:    in  boolean;
    baud_delay_uncertainty: in  time;
    signal tx:              out std_logic
  ) is
  
    constant data_size:    positive                      := data_2_send'length;
    constant data:         std_logic_vector(data_size-1 downto 0) := data_2_send;
    variable seed1, seed2: positive;
    variable rand:         real;
    variable rand_bit:     std_logic_vector (0 downto 0) := "0";
    variable P:            std_logic                     := '0';
    
  begin
    
    tx <= '0';
    
    uniform(seed1, seed2, rand);
    rand_bit := std_logic_vector(to_unsigned(integer(rand), 1));
    if rand_bit = "0" then wait for baud_delay - (baud_delay_uncertainty / 2);
    else wait for baud_delay + (baud_delay_uncertainty / 2);
    end if;
    
    for i in 0 to data_size-1 loop
      tx <= data(i);
      P := P xor data(i);
      uniform(seed1, seed2, rand);
      rand_bit := std_logic_vector(to_unsigned(integer(rand), 1));
      if rand_bit = "0" then wait for baud_delay - (baud_delay_uncertainty / 2);
      else wait for baud_delay + (baud_delay_uncertainty / 2);
      end if;
    end loop;
    
    if inject_parity_error = true then P := not P; end if;
    if parity = 1 then P := not P; end if;
    if parity = 1 or parity = 2 then
      
      tx <= P;
      uniform(seed1, seed2, rand);
      rand_bit := std_logic_vector(to_unsigned(integer(rand), 1));
      
      if ( rand_bit = "0" ) then wait for baud_delay - (baud_delay_uncertainty / 2);
      else wait for baud_delay + (baud_delay_uncertainty / 2);
      end if;
    
    end if;
    
    tx <= '1';
    uniform(seed1, seed2, rand);
    rand_bit := std_logic_vector(to_unsigned(integer(rand), 1));
    
    if rand_bit = "0" then wait for baud_delay - (baud_delay_uncertainty / 2);
    else wait for baud_delay + (baud_delay_uncertainty / 2);
    end if;
    
  
  end procedure;

  procedure rec_uart_frame(
    baud_delay:             in  time;
    parity:                 in  integer range 0 to 2;   -- 0: no parity; 1: odd, 2: even
    timeout:                in  time;
    signal rx:              in  std_logic;
    signal parity_error:    out boolean;
    signal data_received:   out std_logic_vector
  ) is

    constant data_size:    positive   := data_received'length;
    variable data:         std_logic_vector(data_size-1 downto 0) := (others => 'X');
    variable P:            std_logic  := '0';
  begin

    wait until rx = '0' for timeout;
    if rx /= '0' then
      data_received <= data;
      return;
    end if;

    wait for 1.5 * baud_delay; -- sampling the symbols on the center of the interval
    for i in 0 to data_size-1 loop
      data(i) := rx;
      P := P xor rx;
      wait for baud_delay;
    end loop;

    if parity = 1 then P := not P; end if;

    if parity = 1 or parity = 2 then
      parity_error <= P /= rx;
      wait for baud_delay;
    end if;

    -- stop bit, not checked...
    wait for baud_delay;
    data_received <= data;

    wait for 0 ns;

  end procedure;

  procedure bouncing_button(
    period:                 in  time;
    number:                 in  positive;
    signal button:          out std_logic
  ) is

  begin

    for i in 0 to number-1 loop
      button <= not button;
      wait for period;
    end loop;

  end procedure;


  function seven_segment_decoder(
    data: std_logic_vector(7 downto 0)) 
  return std_logic_vector is
    variable result: std_logic_vector(3 downto 0);
  begin

    case data is
      when B"11000000" => result := X"0";
      when B"11111001" => result := X"1";
      when B"10100100" => result := X"2";
      when B"10110000" => result := X"3";
      when B"10011001" => result := X"4";
      when B"10010010" => result := X"5";
      when B"10000010" => result := X"6";
      when B"11111000" => result := X"7";
      when B"10000000" => result := X"8";
      when B"10010000" => result := X"9";
      when B"10001000" => result := X"A";
      when B"10000011" => result := X"B";
      when B"11000110" => result := X"C";
      when B"10100001" => result := X"D";
      when B"10000110" => result := X"E";
      when B"10001110" => result := X"F";
      when others => result := (others => 'X');
    end case;

    return result;
  end function;

  
end package body verification_package;
----------------------------------------------------------------------------------------------------

